/*****************************************************************************
 * Media Library
 *****************************************************************************
 * Copyright © 2022 Hugo Beauzée-Luyssen, Videolabs, VideoLAN
 *
 * Authors: Hugo Beauzée-Luyssen <hugo@beauzee.fr>
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2.1 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin Street, Fifth Floor, Boston MA 02110-1301, USA.
 *****************************************************************************/

#pragma once

#include "medialibrary/ICacher.h"
#include "compat/Mutex.h"
#include "compat/ConditionVariable.h"

namespace VLC
{
class MediaPlayer;
}

namespace medialibrary
{

class LibvlcCacher : public ICacher
{
public:
    LibvlcCacher();
    virtual bool cache( const std::string& inputMrl,
                        const std::string& outputPath ) override;
    virtual void interrupt() override;

private:
    void stop();

private:
    compat::Mutex m_mutex;
    compat::ConditionVariable m_cond;
    VLC::MediaPlayer* m_currentMp;
};

}
